#include <time.h>
#include <sys/time.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <iostream.h>

#include "ControlCentre.h"
#include "DBElement.h"
#include "DB.h"
#include "Buffer.h"
#include "GnutellaPacket.h"
#include "Connection.h"
#include "GnutellaConnection.h"
#include "Host.h"
#include "Networker.h"

Networker::Networker(ControlCentre *newMCC)
{
	MCC			= newMCC;
	poll			= 0;
	listening		= 0;
	pendingConnections	= new DB(MCC->get_MAX_CONNECTIONS() / 2);
	pendingPackets		= new DB(MCC->get_MAX_CONNECTIONS() + 1);
}

Networker::~Networker()
{
	stopListening();
}

GnutellaPacket *Networker::waitTraffic(DB *connections)
{
	GnutellaPacket		*pstart;
	GnutellaPacket		*pprv;
	GnutellaPacket		*p;
	GnutellaConnection	*c;
	char			*buf;
	int			i;
	unsigned int		len;

	readAllConnections(connections);

	/*
	 * See what pending packets are ready
	 */

	pstart	= 0;
	for (i = 0; i < pendingPackets->getSize(); i++)
	{
		if (!(p = (GnutellaPacket *)pendingPackets->findElement(i)))
			continue;

		if (!(c = (GnutellaConnection *)connections->findElement(p->getConnection()))) 		{
			pendingPackets->deleteElement(i);
			continue;
		}

		if (c->getState() != CON_STATE_ACTIVE)
			continue;

		if (c->getInBufferBytes() >= p->getDataLength())
		{
			len	= p->getDataLength();
			if (!((buf = c->getInBufferp(len))))
			{
				MCC->addError("BUG : Networker : couldn't get pending packet body");
				continue;
			}

			pendingPackets->removeElement(i);
			if (p->setBodyDirect(buf, len) != len)
			{
				MCC->addError("BUG : Networker : couldn't set pending packet body");
				p->setType(FUNC_INVALID);
				pprv	= pstart;
				pstart	= p;
				pstart->setLeft(pprv);
				c->punchRecvd();
				c->setLock(0);
				continue;
			}
			pprv	= pstart;
			pstart	= p;
			pstart->setLeft(pprv);
			c->punchRecvd();
			c->setLock(0);
		}
	}

	/*
	 * Create new pending Packets. Any ready packets are added to the list.
	 */
	for (i = 0; i < connections->getSize(); i++)
	{
		if (!(c = (GnutellaConnection *)connections->findElement(i)))
			continue;

		if ((c->getLock()) || (c->getInBufferBytes() <= 0))
			continue;

		if (c->getInBufferBytes() >= GNUTELLA_HEADER_LENGTH)
			p	= new GnutellaPacket(c->getEntry(), MCC);
		else
			continue;

		/*
		 * There is enough data to create a packet
		 */
		if (!((buf = c->getInBufferp(GNUTELLA_HEADER_LENGTH))))
		{
			/*
			 * This shouldn't be possible
			 */
			MCC->addError("BUG : Networker : Error reading from buffer (0),"
					" can't create packet!!");
 			delete p;
			continue;
		}
		p->readHeaderFromString((unsigned char *)buf);
		delete buf;

		if (p->getDataLength() > MAX_BODY_LENGTH)
		{
			/*
			 * Someone is pulling our leg / bandwidth
			 * This packet is not sane. There's no point in reading more from
			 * this connection, an oversized packet doesn't even fit in the buffer.
 			 */
			p->setBodyDirect(0, 0);
			p->setType(FUNC_INVALID);
			pprv	= pstart;
			pstart	= p;
			pstart->setLeft(pprv);
			c->punchRecvd();
			continue;
		}
		
		/*
		 * Nate Riffe discovered that on slow machines, data eventually
		 * clogs up on incoming connections. His proposed fix was to read
		 * all available packets from a connection, in stead of just the first.
		 * Let's see if that helps. In order to continue reading on a connection
		 * that delivered a complete packet, I added a i--; in Cases #1 and #2.
		 */

		/*
		 * Case #1 No body.
		 */
		if (!p->getDataLength())
		{
			pprv	= pstart;
			pstart	= p;
			pstart->setLeft(pprv);
			c->punchRecvd();
			i--;
			continue;
		}

		/*
		 * Case #2 Body is complete
		 */
		if (p->getDataLength() <= c->getInBufferBytes())
		{
			len	= p->getDataLength();
			if (!((buf	= c->getInBufferp(len))))
			{
				/*
				 * This shouldn't be possible
				 */
				MCC->addError("BUG : Networker : Error reading from buffer (2),"
						" can't create packet");
	 			delete p;
				continue;
			}
			if (p->setBodyDirect(buf, len) != len)
			{
				/*
				 * This shouldn't be possible
				 */
				MCC->addError("BUG : Networker : Can't set packet body");
				delete p;
				free(buf);
				continue;
			}
			pprv	= pstart;
			pstart	= p;
			pstart->setLeft(pprv);
			c->punchRecvd();
			i--;
			continue;
		}

		/*
		 * Case #3 Body is incomplete. lock connection 
		 * and store pending packet.
		 */
		if (pendingPackets->isFull() ||
				(pendingPackets->storeElement(p) == -1) )
		{
			/*
			 * Now we're screwed!
			 * We've read half the packet, and can't leave it
			 * anywhere ! Solution : Read rest of packet, then
			 * drop it.
			 */
			MCC->addError("Can't store pending packet!!!");
			c->skipBytes(p->getDataLength());
			delete p;
		}
		else
			c->setLock(1);
	}
	return pstart;
}

int Networker::doSend(DB *connections)
{
	Connection	*c;
	fd_set		out_set;
	struct timeval	tv;
	int		max;
	int		i;
	int		con;

	max		= 0;
	tv.tv_sec	= 0;
	tv.tv_usec	= 1;
        FD_ZERO( &out_set );

	for (i = 0; i < connections->getSize(); i++)
		if ((c = (Connection *)connections->findElement(i)))
		{
			if (	(c->getState() == CON_STATE_ACTIVE)	&&
				(c->getType()  == CON_TYPE_GNUTELLA)	)
			{
				con	= c->getConnection();
				FD_SET(con, &out_set);
				if (con > max)
					max	= con;
			}
		}

	if (!max)
		return 0;

	if (select (max + 1, 0, &out_set, 0, &tv) <= 0)
		return 0;

	for (i = 0; i < connections->getSize(); ++i)
		if ((c = (Connection *)connections->findElement(i)))
		{
			con	= c->getConnection();
			if (FD_ISSET(con, &out_set))
				c->bufferSend();
		}

	return 1;
}

void Networker::readAllConnections(DB *connections)
{
	Connection		*c;
        fd_set 			in_set;
	int			max;
	int			con;
	int			end;
	int			active;
	int			i;
	int			first;
	struct timeval		tv;

	max		= 0;
	tv.tv_sec	= 0;
	tv.tv_usec	= 1;
        FD_ZERO( &in_set  );
	
	for (i = 0; i < connections->getSize(); i++)
		if ((c = (Connection *)connections->findElement(i)))
		{
			if ( 	(c->getState() == CON_STATE_ACTIVE) &&
				(c->getType()  == CON_TYPE_GNUTELLA) 	)
			{
				con	= c->getConnection();
				FD_SET(con, &in_set);
				if (con > max)
					max	= con;
			}
		}

	if (!max)
		return;

	/*
	 * Check events on connections
	 */

	active	= 0;
	end	= poll;
	if (poll >= connections->getSize())
		poll	= 0;

	if (select (max + 1, &in_set, 0, 0, &tv) <= 0)
		return;

	first	= 1;
	while ( ((poll != end) || first) && (active <= POLL_HOSTS) )
	{
		first	= 0;
		if ((c = (Connection *)connections->findElement(poll)))
		{
			if (c->getState() != CON_STATE_ACTIVE)
			{
				if (++poll == connections->getSize())
					poll	-= connections->getSize();
				continue;
			}

			con	= c->getConnection();

			/*
			 * handle connections if case of incoming traffic
			 */
			if (FD_ISSET(con, &in_set))
				if (!c->read())
				{
					c->closeConnection();
					connections->deleteElement(poll);
				}
		}
		if (++poll == connections->getSize())
			poll	-= connections->getSize();
	}
}

Connection *Networker::waitConnection(DB *hosts)
{
	int			max;
	int			i;
	struct timeval		null_time;
        fd_set 			in_set;
	char			identifier[50];
	Connection		*c;
	Connection		*prv;
	Connection		*first;

	doListen(hosts);

	max	= 0;
	FD_ZERO(&in_set);
	for (i = 0; i < pendingConnections->getSize(); i++)
		if ((c = (Connection *)pendingConnections->findElement(i)))
		{
			FD_SET(c->getConnection(), &in_set);
			if (c->getConnection() > max)
				max	= c->getConnection();
		}
	
	null_time.tv_sec	= 0;
	null_time.tv_usec	= 0;
	if (select(max + 1, &in_set, 0, 0, &null_time) <= 0)
		return 0;

	first		= 0;
	for (i = 0; i < pendingConnections->getSize(); i++)
		if ((c = (Connection *)pendingConnections->findElement(i)))
		{
			if (FD_ISSET(c->getConnection(), &in_set))
			{
				c->read();
				if (c->getInBufferBytes() < 10)
				{
					if (c->getAge()	> 30)
					{
						c->closeConnection();
						pendingConnections->deleteElement(i);
					}
					continue;
				}
					
				c->read();
				c->peekInBuffer(identifier, 49);
				identifier[49]	= 0;

				if (!strncmp(identifier, "GNUTELLA CONNECT/0.4\n\n",
					strlen("GNUTELLA CONNECT/0.4\n\n")))
				{
					if (c->getInBufferBytes() < strlen("GNUTELLA CONNECT/0.4\n\n"))
					{
						if (c->getAge()	> 30)
						{
							c->closeConnection();
							pendingConnections->deleteElement(i);
						}
						continue;
					}
					c->setType(CON_TYPE_GNUTELLA);
				}
				else if (!strncmp(identifier, "GET /get/",
						strlen("GET /get/")))
				{
					if (c->getInBufferBytes() < 12)
					{
						if (c->getAge()	> 30)
						{
							c->closeConnection();
							pendingConnections->deleteElement(i);
						}
						continue;
					}
					c->setType(CON_TYPE_UPLOAD);
				}
				else if (!strncmp(identifier, "GIV ",
						strlen("GIV ")))
				{
					if (c->getInBufferBytes() < 40)
					{
						if (c->getAge()	> 30)
						{
							c->closeConnection();
							pendingConnections->deleteElement(i);
						}
						continue;
					}
					c->setType(CON_TYPE_PUSH_IN);
				}
				else
				{
					c->closeConnection();
					pendingConnections->deleteElement(i);
					c	= 0;
				}
				if (c)
				{
					c->setState(CON_STATE_CONNECTED);
					pendingConnections->removeElement(i);
					prv	= first;
					first	= c;
					first->setLeft(prv);
				}
			}
		}
	return first;
}

void Networker::doListen(DB *hosts)
{
	Connection		*c;
	Host			*h;
	struct sockaddr_in	sa;
	int			con;
	int			HIndex;
	unsigned int		len;
	struct timeval		tm;
	fd_set			in_set;
	
	if ( (!isListening()) || hosts->isFull() ||
				pendingConnections->isFull() )
		return;

	tm.tv_sec	= 0;
	tm.tv_usec	= 1;
	FD_ZERO(&in_set);
	FD_SET(sock, &in_set);
	if (select(sock + 1, &in_set, NULL, NULL, &tm) <= 0)
		return;

	len	= sizeof(sa);
	if ((con = accept(sock, (struct sockaddr *)&sa, &len)) < 0)
		return;
	h	= new Host(sa.sin_addr.s_addr, ntohs(sa.sin_port));
	h->setReal(0);

	if (((HIndex = hosts->containsElement(h))) != ENTRY_INVALID)
		delete h;
	else if (((HIndex = hosts->storeElement(h))) == ENTRY_INVALID)
	{
		close (con);
		delete h;
		return;
	}
	c	= new Connection(con, HIndex);
	if (pendingConnections->storeElement(c) == ENTRY_INVALID)
	{
		c->closeConnection();
		delete c;
		return;
	}
}

int Networker::isListening()
{
	return listening;
}

int Networker::startListening()
{
	const int		reuseaddr	= 1;
	struct sockaddr_in	sa;


	if (listening)
		return 0;

	sa.sin_family		= AF_INET;
	sa.sin_addr.s_addr	= htonl(INADDR_ANY);
	sa.sin_port     	= htons( MCC->get_PORT() );

	if ((sock = socket(PF_INET, SOCK_STREAM, 0)) < 0)
	{
		MCC->addError("ERROR OPENING SOCKET!!!");
		return 0;
	}

	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &reuseaddr, 
			sizeof(reuseaddr));

	if ( bind(sock, (struct sockaddr *)&sa, sizeof(sa)) < 0)
	{
		MCC->addError("ERROR BINDING SOCKET!!!");
		close (sock);
		return 0;
	}

	listen(sock, 3);
	listening		= 1;

	return 1;
}

int Networker::stopListening()
{
	if (!listening)
		return 0;

	close(sock);
	listening	= 0;
	return 1;
}

/*
Networker::
Networker::
*/
